import contextlib
import functools

from collections import OrderedDict
from typing import Callable, ContextManager, Dict, List, Optional, Set, Union

import attr

from snowflake.connector import SnowflakeConnection
from snowflake.connector.cursor import SnowflakeCursor

from dbnd_snowflake import log_snowflake_resource_usage, log_snowflake_table
from dbnd_snowflake.extract_sql_query import TableTargetOperation, extract_from_sql
from dbnd_snowflake.sql_utils import is_sql_crud_command
from targets.connections import build_conn_path


@attr.s
class SessionData(object):
    query_ids = attr.ib(factory=list)

    def is_empty(self):
        # if we have tables but no query ids?
        # return not self.query_ids and not self.tables  ???
        return not self.query_ids

    def add_query(self, query_id):
        if query_id:
            self.query_ids.append(query_id)


@attr.s
class ConnectionData(object):
    schema = attr.ib()
    database = attr.ib()
    connection = attr.ib(default=None)  # type: SnowflakeConnection
    tables_ops = attr.ib(factory=set)  # type: Set[TableTargetOperation]
    session_data = attr.ib(factory=OrderedDict)  # type: Dict[int, SessionData]
    connection_number = attr.ib(default=0)  # type: int

    def get_session_data(self, session_id):
        # type: (int) -> SessionData
        # we need this before `if` to correctly track nested track() calls
        if session_id not in self.session_data:
            self.session_data[session_id] = SessionData()
        return self.session_data[session_id]

    def reset_session_queries(self):
        self.session_data = OrderedDict()

    def reset_tables(self):
        self.tables_ops = set()

    def get_all_session_queries(self):
        return OrderedDict(
            [
                (session_id, session_data.query_ids)
                for session_id, session_data in self.session_data.items()
            ]
        )

    def get_connection_path(self):
        return build_conn_path(
            conn_type="snowflake",
            hostname=self.connection.host,
            port=self.connection.port,
            path="/".join(filter(None, [self.database, self.schema])),
        )


# track connection either when specific connection is closed or when leaving the context:
# Example 1:
# with snowflake_query_tracker():
#   with new_snowflake_connection() as conn1:
#     ... use conn1 ...
#     <if connection is closing - will track here>
#   conn2 = another_snowflake_connection():
#   ... use conn2 ... (wasn't closed)
# <track all not closed connections>
# Example 2:
# with snowflake_conn() as conn3:
#    with snowflake_query_tracker():
#      ... use conn3 ...
#       <track connection usage>
class SnowflakeQueryTracker(object):
    def __init__(
        self,
        track_connection_handler: Callable[[ConnectionData], None] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
    ):
        self._connection_data = OrderedDict()  # type: Dict[int, ConnectionData]
        self._suspended = False
        self._track_connection_handler = track_connection_handler

        self.database = database
        self.schema = schema

        self.last_session_id = None
        self.last_session_query_ids = []

        self.connection_counter = 0

    def __enter__(self):
        self.patch_snowflake()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            # track all connections data when leaving the tracker scope (example 2)
            for cd in self._connection_data.values():
                self.track_connection_data(cd)
        finally:
            self.unpatch_snowflake()

    @contextlib.contextmanager
    def track_execute(self, cursor, command, *args, **kwargs):
        # type: (SnowflakeCursor, str, ..., ...) -> None

        if self._suspended or not is_sql_crud_command(command):
            yield
            return

        connection_data = self._get_connection_data(cursor.connection)

        session_id = cursor.connection.session_id
        session_data = connection_data.get_session_data(session_id)

        conn_path = connection_data.get_connection_path()
        tables_ops = extract_from_sql(conn_path, command)
        connection_data.tables_ops.update(tables_ops)

        try:
            yield
        finally:
            query_id = cursor.sfqid
            if query_id:
                session_data.add_query(query_id)

                # this is needed only to support self.get_last_session_with_query_id()
                # for backward compatibility, remove some day (2020/11/26)
                if self.last_session_id != session_id:
                    self.last_session_id = session_id
                    self.last_session_query_ids = []
                self.last_session_query_ids.append(query_id)

    def _get_connection_data(self, connection):
        # type: (SnowflakeConnection) -> ConnectionData
        conn_id = id(connection)
        if conn_id not in self._connection_data:
            self._connection_data[conn_id] = ConnectionData(
                database=connection.database or self.database,
                schema=connection.schema or self.schema,
                connection=connection,
                connection_number=self.connection_counter,
            )
            self.connection_counter += 1
        return self._connection_data[conn_id]

    def get_all_tables(self):
        tables = {
            table_op.name
            for connection_data in self._connection_data.values()
            for table_op in connection_data.tables_ops
        }
        return tables

    def get_all_session_queries(self):
        session_queries = OrderedDict()
        for connection_data in self._connection_data.values():  # type: ConnectionData
            session_queries.update(connection_data.get_all_session_queries())
        return session_queries

    def patch_snowflake(self):
        self._patch_cursor_execute()
        self._patch_connection_close()

    def _patch_cursor_execute(self):
        if hasattr(SnowflakeCursor.execute, "__dbnd_patched__"):
            # already patched
            # warn here ?
            return

        original = SnowflakeCursor.execute

        @functools.wraps(original)
        def snowflake_cursor_execute(cursor_self, *args, **kwargs):
            with self.track_execute(cursor_self, *args, **kwargs):
                return original(cursor_self, *args, **kwargs)

        snowflake_cursor_execute.__dbnd_patched__ = original
        SnowflakeCursor.execute = snowflake_cursor_execute

    def _patch_connection_close(self):
        if hasattr(SnowflakeConnection.close, "__dbnd_patched__"):
            # already patched
            # warn here ?
            return

        original = SnowflakeConnection.close

        @functools.wraps(original)
        def snowflake_connection_close(connection_self, *args, **kwargs):
            # track connection before closing it (Example 1)
            self.track_connection_data(self._connection_data.get(id(connection_self)))
            return original(connection_self, *args, **kwargs)

        snowflake_connection_close.__dbnd_patched__ = original  # or "= self" ?
        SnowflakeConnection.close = snowflake_connection_close

    def unpatch_snowflake(self):
        if hasattr(SnowflakeCursor.execute, "__dbnd_patched__"):
            SnowflakeCursor.execute = SnowflakeCursor.execute.__dbnd_patched__

        if hasattr(SnowflakeConnection.close, "__dbnd_patched__"):
            SnowflakeConnection.close = SnowflakeConnection.close.__dbnd_patched__

    def get_last_session_with_query_id(self, many):
        if not self.last_session_query_ids:
            return None, ([] if many else None)

        return (
            self.last_session_id,
            (self.last_session_query_ids if many else self.last_session_query_ids[-1]),
        )

    @contextlib.contextmanager
    def suspend(self):
        self._suspended = True
        yield
        self._suspended = False

    def track_connection_data(self, conn_data: ConnectionData):
        if not self._track_connection_handler or not conn_data:
            return
        if conn_data.connection.is_closed():
            # already closed, cannot proceed (and probably already tracked)
            return

        with self.suspend():
            self._track_connection_handler(conn_data)


def log_all_snowflake_resource_usage(cd, reset=True, **kwargs):
    # type: (ConnectionData, bool, ...) -> Optional[Dict[int, List[str]]]
    # copy to prevent resource usage queries to appear in current history
    history = cd.get_all_session_queries()
    for session_id, query_ids in history.items():
        log_snowflake_resource_usage(
            database=cd.database,
            connection_string=cd.connection,
            session_id=session_id,
            query_ids=query_ids,
            key=f"snowflake_query.{cd.connection_number}",
            **kwargs,
        )
    if reset:
        cd.reset_session_queries()
    return history


def log_all_snowflake_tables(cd, tables, reset=True, **kwargs):
    # type: (ConnectionData, Union[bool, List[str]], bool, ...) -> Optional[List[str]]
    """

    :param cd: ConnectionData object
    :param tables: if True - will log automatically discovered tables; if list of strings - will log only mentioned tables
    :param with_preview:
    :param with_schema:
    :param raise_on_error:
    :param reset: should reset the ConnectionData.tables once done
    :return:
    """
    if tables is True:
        tables = [table_op.name for table_op in cd.tables_ops]
    else:
        # don't reset if explicit tables provided
        reset = False

    if tables:
        for table in tables:
            parts = table.split(".")
            table_name = parts.pop()
            schema = parts.pop() if parts else cd.schema
            database = parts.pop() if parts else cd.database

            log_snowflake_table(
                table_name=table_name,
                schema=schema,
                database=database,
                connection_string=cd.connection,
                **kwargs,
            )
    if reset:
        cd.reset_tables()
    return tables


# TODO: better way of passing log_tables_*/log_resource_* parameters
#  maybe by passing partial log_snowflake_resource_usage()/log_snowflake_table()
#  with preset configuration values, or using SnowflakeConfig/explicit configs
@contextlib.contextmanager
def snowflake_query_tracker(
    log_resource_usage: bool = True,
    log_tables: Union[bool, List[str]] = True,
    log_tables_with_preview: Optional[bool] = None,
    log_tables_with_schema: Optional[bool] = None,
    log_resource_history_window: float = 15,
    log_resource_query_history_result_limit: Optional[int] = None,
    log_resource_delay: int = 1,
    log_resource_retries: int = 3,
    log_resource_retry_pause: float = 0,
    raise_on_error: bool = False,
    database=None,
    schema=None,
) -> ContextManager[SnowflakeQueryTracker]:
    """

    :param log_resource_usage: bool whether to track resource usage
    :param log_tables: either explicit list of tables to log or True to log all automatically extracted tables from queries
    :param log_tables_with_preview: to log tables preview
    :param log_tables_with_schema: to log tables schemas
    :param log_resource_history_window: How deep to search into QUERY_HISTORY. Set in minutes
    :param log_resource_query_history_result_limit: Passed through directly to QUERY_HISTORY search function as `RESULT_LIMIT` param
    :param log_resource_delay: Initial delay before looking in QUERY_HISTORY.
        Metadata can appear there with some delay. Use this param for fine tuning
    :param log_resource_retries: How much times to search in QUERY_HISTORY.
        Each time search is widened by increasing `RESULT_LIMIT` param.
    :param log_resource_retry_pause: Set number of seconds to pause before next retry.
    :param raise_on_error: By default all exceptions are muted so your task success status
        is not affected by errors in tracking. Set to true to re-raise all exceptions.
    :param database: database to use for resource usage
    :param schema: schema to use
    """

    def _track_connection(connection_data):
        # type: (ConnectionData) -> None
        if log_resource_usage:
            log_all_snowflake_resource_usage(
                connection_data,
                history_window=log_resource_history_window,
                query_history_result_limit=log_resource_query_history_result_limit,
                delay=log_resource_delay,
                retries=log_resource_retries,
                retry_pause=log_resource_retry_pause,
                raise_on_error=raise_on_error,
                reset=True,
            )
        if log_tables:
            log_all_snowflake_tables(
                connection_data,
                tables=log_tables,
                with_preview=log_tables_with_preview,
                with_schema=log_tables_with_schema,
                raise_on_error=raise_on_error,
                reset=True,
            )

    with SnowflakeQueryTracker(
        track_connection_handler=_track_connection, database=database, schema=schema,
    ) as st:
        yield st



from dbnd import parameter
from dbnd._core.task import Config


class SnowflakeConfig(Config):
    """dbnd-snowflake configuration"""

    _conf__task_family = "snowflake"

    table_preview_rows = parameter(
        default=20,
        description="How many (first) rows to include in Snowflake table preview",
    )[int]

    query_history_result_limit = parameter(
        default=100,
        description="Default value to use when searching for a query resources in QUERY_HISTORY",
    )[int]

    query_history_result_limit_max_value = parameter(
        default=10000,
        description="RESULT_LIMIT param max value accepted by QUERY_HISTORY() function. Hard limit imposed by Snowflake.",
    )[int]

    query_history_end_time_range_start = parameter(
        default=15,
        description="QUERY_HISTORY END_TIME_RANGE_START param. Minutes before now.",
    )[int]

    query_history_end_time_range_end = parameter(
        default=0,
        description="QUERY_HISTORY END_TIME_RANGE_END param. Minutes before now.",
    )[int]



import logging
import typing

from typing import Optional

import attr

from dbnd_snowflake.snowflake_config import SnowflakeConfig
from targets.value_meta import ValueMeta, ValueMetaConf
from targets.values import register_value_type
from targets.values.builtins_values import DataValueType


if typing.TYPE_CHECKING:
    from dbnd_snowflake.snowflake_controller import SnowflakeController

logger = logging.getLogger(__name__)


def _strip_quotes(v):
    # type: (str) -> str
    return v.strip('"') if v else v


@attr.s
class SnowflakeTable(object):
    snowflake_ctrl = attr.ib()  # type: SnowflakeController
    database = attr.ib(converter=_strip_quotes)  # type: str
    schema = attr.ib(converter=_strip_quotes)  # type: str
    table_name = attr.ib(converter=_strip_quotes)  # type: str
    preview_rows = attr.ib(default=20)  # type: int

    @property
    def db_with_schema(self):
        if self.schema:
            return "{0}.{1}".format(self.database, self.schema)
        return self.database

    @property
    def database_q(self):
        return '"{}"'.format(self.database) if self.database else self.database

    @property
    def schema_q(self):
        return '"{}"'.format(self.schema) if self.schema else self.schema

    @property
    def table_name_q(self):
        return '"{}"'.format(self.table_name) if self.table_name else self.table_name

    @property
    def db_with_schema_q(self):
        if self.schema:
            return "{0}.{1}".format(self.database_q, self.schema_q)
        return self.database

    def __str__(self):
        return "{0}.{1}".format(self.db_with_schema, self.table_name)

    @classmethod
    def from_table(cls, snowflake_ctrl, table, config=None):
        # type: (SnowflakeController, str, Optional[SnowflakeConfig]) -> SnowflakeTable
        if config is None:
            config = SnowflakeConfig.current()
        preview_rows = config.table_preview_rows

        # this case is very simple, we need to extract from the controller if there is a missing part
        rest, _, table_name = table.rpartition(".")
        database, _, schema = rest.rpartition(".")

        return cls(snowflake_ctrl, database, schema, table_name, preview_rows)


class SnowflakeTableValueType(DataValueType):
    type = SnowflakeTable
    type_str = "SnowflakeTable"
    support_merge = False

    config_name = "snowflake_table"

    is_lazy_evaluated = True

    def to_signature(self, x):
        # type: (SnowflakeTable) -> str
        return "{0.snowflake_ctrl}/{0.db_with_schema}/{0.table_name}".format(x)

    def get_value_meta(self, value, meta_conf):
        # type: (SnowflakeTable, ValueMetaConf) -> ValueMeta
        data_schema = {}
        data_preview = data_dimensions = None
        if meta_conf.log_preview:
            data_preview = value.snowflake_ctrl.to_preview(value)

        if meta_conf.log_schema:
            data_schema = {
                "type": self.type_str,
                "column_types": value.snowflake_ctrl.get_column_types(value),
            }

        if meta_conf.log_size:
            dimensions = value.snowflake_ctrl.get_dimensions(value)
            data_dimensions = [dimensions["rows"], dimensions["cols"]]
            data_schema["size.bytes"] = dimensions["bytes"]

        # currently stats and histogram are not supported
        stats, histograms = {}, {}
        hist_sys_metrics = None

        return ValueMeta(
            value_preview=data_preview,
            data_dimensions=data_dimensions,
            data_schema=data_schema,
            data_hash=str(hash(self.to_signature(value))),
            descriptive_stats=stats,
            histogram_system_metrics=hist_sys_metrics,
            histograms=histograms,
        )


register_value_type(SnowflakeTableValueType())



import re


def try_extract_tables(command):
    if not command:
        return

    # TODO: this is not the right way to do it! it's not good idea to
    #  parse SQL with regex! If this direction is good, it's better to
    #  use SQL parser library, for example https://github.com/andialbrecht/sqlparse

    # catch all usages:
    # select from X ..
    # .. join X ...
    # update X ...
    # insert into X ...
    # delete from X ...
    tables = re.findall(
        r"\b(?:FROM|JOIN|UPDATE|INSERT\s+(?:\bINTO\b)?|DELETE\s+(?:\bFROM\b)?)[\s(]+([\"a-zA-Z._\d]+)",
        command,
        re.IGNORECASE,
    )

    # if code includes "with" statement, we should remove those tmp "tables", like:
    # WITH A as (select * from ...), B as (select * from ...) select ...

    # 1. first remove all inner select clauses to be able to catch A and B:
    # from query above, will keep only WITH A as , B as select ...
    cmd = command
    for i in range(20):
        new_cmd = re.sub(r"\([^()]*\)", "", cmd)
        if len(new_cmd) == len(cmd):
            break
        cmd = new_cmd

    # doesn't support recursive with (with recursive a(n) as ())
    # 2. extract only with-related part (WITH A as , B as)
    with_stmt_match = re.search(
        r"\bWITH\s+(([a-zA-Z._\d]+)\s+as\s*,?\s*)+", cmd, re.IGNORECASE,
    )
    if with_stmt_match:
        # 3. extract all "tables" names (A , B)
        with_stmt = re.findall(
            r"\b([a-zA-Z._\d]+)\s+as\b", with_stmt_match[0], re.IGNORECASE
        )
    else:
        with_stmt = []
    tables = [
        x
        for x in (set(tables) - set(with_stmt))
        if x.upper() not in ("SELECT", "TABLE")
    ]
    return tables


def is_sql_crud_command(command):
    # if all commands are ALTER (like ALTER SESSION or ALTER TABLE) or CREATE
    # (like CREATE TABLE) - don't track. Basically we'd want to track only queries
    # which retrieve or modify data
    # warning: ";" could be part of the some internal "where a like '%;%'"
    # but this shouldn't affect the correctness of the function
    for cmd in command.split(";"):
        if re.search(
            r"^\s*\b(SELECT|INSERT|DELETE|UPDATE|MERGE|WITH)\b", cmd, re.IGNORECASE
        ):
            return True
    return False



import logging

from typing import List, Optional, Union

from snowflake.connector import SnowflakeConnection

from dbnd import log_duration, log_metrics
from dbnd_snowflake.snowflake_config import SnowflakeConfig
from dbnd_snowflake.snowflake_controller import SnowflakeController, SnowflakeError


logger = logging.getLogger(__name__)


def log_snowflake_resource_usage(
    database: str,
    connection_string: Union[str, SnowflakeConnection],
    query_ids: List[str],
    session_id: Optional[int] = None,
    key: str = "snowflake_query",
    history_window: float = 15,
    query_history_result_limit: Optional[int] = None,
    delay: int = 1,
    retries: int = 3,
    retry_pause: float = 0,
    raise_on_error: bool = False,
) -> None:
    """
    Search for a query previously executed by Snowflake in it's QUERY_HISTORY and log cpu time,
    run time, disk read, and other resources.

    Query's metadata can appear in QUERY_HISTORY with a lag up to 45 minutes.

    :param database: Name of the database query was issued to.
    :param connection_string: Snowflake connection string to use or existing SnowflakeConnection object.
    :param query_ids: Supply a list of `query_id` generated by Snowflake for search in QUERY_HISTORY.
    :param session_id: Supply `session_id` generated by Snowflake for more efficient search in QUERY_HISTORY.
    :param key: Override it if you call this function twice or more within the same task/Airflow Operator
    :param history_window: How deep to search into QUERY_HISTORY. Set in minutes
    :param query_history_result_limit: Passed through directly to QUERY_HISTORY search function as `RESULT_LIMIT` param
    :param delay: Initial delay before looking in QUERY_HISTORY.
        Metadata can appear there with some delay. Use this param for fine tuning
    :param retries: How much times to search in QUERY_HISTORY.
        Each time search is widened by increasing `RESULT_LIMIT` param.
    :param raise_on_error: By default all exceptions are muted so your task success status
        is not affected by errors in tracking. Set to true to re-raise all exceptions.
    :param retry_pause: Set number of seconds to pause before next retry.
    """

    with SnowflakeController(connection_string) as snowflake_ctrl:
        snowflake_config = SnowflakeConfig()
        if query_history_result_limit is None:
            query_history_result_limit = snowflake_config.query_history_result_limit

        if not all(query_id for query_id in query_ids):
            error_msg = f"query_ids cannot be empty. You supplied: {query_ids}"
            if raise_on_error:
                raise SnowflakeError(error_msg)
            else:
                logger.error(error_msg)
                return

        metrics_to_log = {}
        # XXX: Do we actually need log_duration?
        with log_duration("log_snowflake_resource_usage__time_seconds", "system"):
            for i, query_id in enumerate(query_ids):
                query_key = f"{key}.{i}" if len(query_ids) > 1 else key
                metrics_to_log.update(
                    snowflake_ctrl.get_resource_usage(
                        database,
                        query_id,
                        session_id,
                        query_key,
                        history_window,
                        query_history_result_limit,
                        delay,
                        retries,
                        retry_pause,
                        raise_on_error,
                        snowflake_config,
                    )
                )

        log_metrics(metrics_to_log, source="user")



import logging

import dbnd


logger = logging.getLogger(__name__)


@dbnd.hookimpl
def dbnd_setup_plugin():
    from dbnd import register_config_cls
    from dbnd_snowflake.snowflake_config import SnowflakeConfig

    register_config_cls(SnowflakeConfig)
